mod users;
mod webfinger;

use core::fmt;
use std::task::{Context, Poll};

use axum::{
    Router,
    http::{StatusCode, Uri},
    routing::get,
};
use log::{info, trace};
use tower_layer::Layer;
use tower_service::Service;
use users::users_get;
use webfinger::webfinger_get;

const HOST: &'static str = "testhost.com";
const USER: &'static str = "test";

#[tokio::main]
async fn main() {
    env_logger::builder().init();

    let app = Router::new()
        .route("/", get(root))
        .route("/.well_known/webfinger", get(webfinger_get))
        .route("/users/{user}", get(users_get))
        .fallback(fallback)
        .layer(TraceLayer);

    let listener = tokio::net::TcpListener::bind("0.0.0.0:8000").await.unwrap();
    axum::serve(listener, app).await.unwrap();
}

async fn root() -> &'static str {
    "Hello World!"
}

async fn fallback(uri: Uri) -> (StatusCode, String) {
    info!("Got request for non implemented route: {uri}");
    (StatusCode::NOT_FOUND, format!("No route for {uri}"))
}

#[derive(Clone)]
pub struct TraceLayer;

impl<S> Layer<S> for TraceLayer {
    type Service = TraceService<S>;

    fn layer(&self, service: S) -> Self::Service {
        TraceService { service }
    }
}

// This service implements the Log behavior
#[derive(Clone)]
pub struct TraceService<S> {
    service: S,
}

impl<S, Request> Service<Request> for TraceService<S>
where
    S: Service<Request>,
    Request: fmt::Debug,
{
    type Response = S::Response;
    type Error = S::Error;
    type Future = S::Future;

    fn poll_ready(&mut self, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
        self.service.poll_ready(cx)
    }

    fn call(&mut self, request: Request) -> Self::Future {
        // Insert log statement here or other functionality
        trace!("request = {:?}", request);
        self.service.call(request)
    }
}
